==> welltested-2.1.4.tar.gz: lib/generators/saver.dart <==
/*
 * Created on May 1st 2023
 *
 * Copyright (C) 2023 Welltested AI - All Rights Reserved
 * The code in this file is Intellectutal Property of Welltested AI.
 * and can't be edited, redistributed or used without a valid license of it's product (Welltested)[https://welltested.ai]
 */

import 'dart:io';

import 'package:analyzer/dart/ast/ast.dart';
import 'package:analyzer/dart/element/element.dart';
import 'package:azure_application_insights/azure_application_insights.dart';
import 'package:welltested/src/executors/fixer_executor.dart';
import 'package:welltested/src/executors/unit_test_executor.dart';
import 'package:welltested/src/extensions/annotation_ext.dart';
import 'package:welltested/src/extensions/class_element_extension.dart';
import 'package:welltested/src/extensions/extension_element_ext.dart';
import 'package:welltested/src/extensions/test_request_extension.dart';
import 'package:welltested/src/models/pubspec_data.dart';
import 'package:welltested/src/models/test_run_info.dart';
import 'package:welltested/src/repositories/dart_server_repository.dart';
import 'package:welltested/src/repositories/generations_repository.dart';
import 'package:welltested/src/utilities/cache_manager.dart';
import 'package:welltested/src/utilities/logging.dart';
import 'package:welltested/src/utilities/path_utils.dart';
import 'package:welltested/src/utilities/telemetry_client.dart';

class WelltestedSaver {
  final DartServerRepository dartServerRepository;
  final CacheManager cacheManager;

  WelltestedSaver(
      {required this.dartServerRepository, required this.cacheManager});
  late final GenerationRepository generationRepository;

  ///Executors
  late final FixerExecutor fixerExecutor;
  late final UnitTestExecutor unitTestExecutor;
  void init() {
    generationRepository = GenerationRepository();
    fixerExecutor = FixerExecutor(dartServerRepository.server);
    unitTestExecutor = UnitTestExecutor(generationRepository);
  }

  Future<void> saveForClass(
      {required ClassElement classElement,
      required ElementAnnotation annotation,
      required CompilationUnit compilationUnit,
      required ClassDeclaration classAstNode,
      required PubspecData pubspecData,
      required File classFile,
      required TestRunInfo testRunInfo}) async {
    final List<String> excludedMethods = annotation.excludeMethods;
    final fileImport =
        'package:${pubspecData.packageName}/${PathUtils.relatetiveLibFilePath(classFile.path)}';

    final methodAstNodes = await classElement.getMethodDeclarations;

    for (int i = 0; i < methodAstNodes.entries.length; i++) {
      final method = methodAstNodes.entries.elementAt(i);
      await _saveTestsForMethod(
          method: method,
          compilationUnit: compilationUnit,
          classAstNode: classAstNode,
          fileImport: fileImport,
          excludedMethods: excludedMethods,
          subjectFile: classFile,
          testRunInfo: testRunInfo,
          className: classElement.displayName,
          methodAstNodes: methodAstNodes);
    }
  }

  Future<void> saveForExtension(
      {required ExtensionElement extElement,
      required ElementAnnotation annotation,
      required CompilationUnit compilationUnit,
      required ExtensionDeclaration extAstNode,
      required PubspecData pubspecData,
      required File extFile,
      required TestRunInfo testRunInfo}) async {
    final List<String> excludedMethods = annotation.excludeMethods;
    final fileImport =
        'package:${pubspecData.packageName}/${PathUtils.relatetiveLibFilePath(extFile.path)}';

    final methodAstNodes = await extElement.getMethodDeclarations;

    for (int i = 0; i < methodAstNodes.entries.length; i++) {
      final method = methodAstNodes.entries.elementAt(i);
      await _saveTestsForMethod(
          method: method,
          compilationUnit: compilationUnit,
          extAstNode: extAstNode,
          fileImport: fileImport,
          excludedMethods: excludedMethods,
          subjectFile: extFile,
          testRunInfo: testRunInfo,
          className: extElement.displayName,
          methodAstNodes: methodAstNodes);
    }
  }

  Future<void> saveForGlobalExecutableElement(
      {required ExecutableElement methodElement,
      required FunctionDeclaration methodAstNode,
      required CompilationUnit compilationUnit,
      required ElementAnnotation annotation,
      required PubspecData pubspecData,
      required File methodFile,
      required TestRunInfo testRunInfo}) async {
    final List<String> excludedMethods = annotation.excludeMethods;
    final fileImport =
        'package:${pubspecData.packageName}/${PathUtils.relatetiveLibFilePath(methodFile.path)}';
    await _saveTestsForMethod(
        method: MapEntry(methodElement.displayName, methodAstNode),
        compilationUnit: compilationUnit,
        fileImport: fileImport,
        excludedMethods: excludedMethods,
        subjectFile: methodFile,
        testRunInfo: testRunInfo);
  }

  Future<void> saveForGlobalMethod(
      {required FunctionElement methodElement,
      required FunctionDeclaration methodAstNode,
      required CompilationUnit compilationUnit,
      required ElementAnnotation annotation,
      required PubspecData pubspecData,
      required File methodFile,
      required TestRunInfo testRunInfo}) async {
    final List<String> excludedMethods = annotation.excludeMethods;
    final fileImport =
        'package:${pubspecData.packageName}/${PathUtils.relatetiveLibFilePath(methodFile.path)}';
    await _saveTestsForMethod(
        method: MapEntry(methodElement.displayName, methodAstNode),
        compilationUnit: compilationUnit,
        fileImport: fileImport,
        excludedMethods: excludedMethods,
        subjectFile: methodFile,
        testRunInfo: testRunInfo);
  }

  Future<void> _saveTestsForMethod({
    required MapEntry<String, Declaration> method,
    required CompilationUnit compilationUnit,
    required String fileImport,
    required List<String> excludedMethods,
    required File subjectFile,
    required TestRunInfo testRunInfo,
    ClassDeclaration? classAstNode,
    ExtensionDeclaration? extAstNode,
    String? className,
    Map<String, MethodDeclaration>? methodAstNodes,
  }) async {
    try {
      final testRequestData = await unitTestExecutor.prepareRequest(
        method: method,
        compilationUnit: compilationUnit,
        classAstNode: classAstNode,
        extensionAstNode: extAstNode,
        fileImport: fileImport,
        className: className,
        methodAstNodes: methodAstNodes,
        excludedMethods: excludedMethods,
      );
      if (testRequestData == null) return;
      final file =
          PathUtils.correspondingPermanentTestFile(method.key, subjectFile);
      if (!(await file.exists())) return;
      final testIsSaved = await cacheManager.checkIfTestIsSaved(
          testRequestData.codeInScrutity(excludeAnnotations: false),
          subjectFile.path,
          testRequestData.functionName,
          className: testRequestData.parentEntityName);
      if (testIsSaved) return; //this test is already saved.
      var hasErrors =
          await fixerExecutor.hasCompilationErrors(file.absolute.path);

      if (hasErrors) {
        wtLog.log(
            "⚠️ File ${PathUtils.relatetiveLibFilePath(file.path)} has errors and will not be updated");
        return;
      }
      final fixedTest = await file.readAsString();
      await unitTestExecutor.updateTest(testRequestData, fixedTest);
      await cacheManager.cacheTest(
          code: testRequestData.codeInScrutity(excludeAnnotations: false),
          codeFilePath: subjectFile.path,
          test: fixedTest,
          testFilePath: file.path,
          functionName: testRequestData.functionName,
          className: testRequestData.parentEntityName,
          saved: true);
      testRunInfo.incrementFromAction(TestGenerationStatus.update);
    } catch (e, stackTrace) {
      wtTelemetry.trackError(
          severity: Severity.error, error: e, stackTrace: stackTrace);
      wtLog.warning("Error saving unit tests for method: ${method.key}");
      wtLog.log(e.toString());
    }
  }
}

==> welltested-2.1.6.tar.gz: lib/generators/saver.dart <==
/*
 * Created on May 1st 2023
 *
 * Copyright (C) 2023 Welltested AI - All Rights Reserved
 * The code in this file is Intellectutal Property of Welltested AI.
 * and can't be edited, redistributed or used without a valid license of it's product (Welltested)[https://welltested.ai]
 */

import 'dart:io';

import 'package:analyzer/dart/ast/ast.dart';
import 'package:analyzer/dart/element/element.dart';
import 'package:azure_application_insights/azure_application_insights.dart';
import 'package:welltested/src/executors/fixer_executor.dart';
import 'package:welltested/src/executors/unit_test_executor.dart';
import 'package:welltested/src/extensions/annotation_ext.dart';
import 'package:welltested/src/extensions/class_element_extension.dart';
import 'package:welltested/src/extensions/extension_element_ext.dart';
import 'package:welltested/src/extensions/test_request_extension.dart';
import 'package:welltested/src/models/pubspec_data.dart';
import 'package:welltested/src/models/test_run_info.dart';
import 'package:welltested/src/repositories/dart_server_repository.dart';
import 'package:welltested/src/repositories/generations_repository.dart';
import 'package:welltested/src/utilities/cache_manager.dart';
import 'package:welltested/src/utilities/logging.dart';
import 'package:welltested/src/utilities/path_utils.dart';
import 'package:welltested/src/utilities/telemetry_client.dart';

class WelltestedSaver {
  final DartServerRepository dartServerRepository;
  final CacheManager cacheManager;

  WelltestedSaver(
      {required this.dartServerRepository, required this.cacheManager});
  late final GenerationRepository generationRepository;

  ///Executors
  late final FixerExecutor fixerExecutor;
  late final UnitTestExecutor unitTestExecutor;
  void init() {
    generationRepository = GenerationRepository();
    fixerExecutor = FixerExecutor(dartServerRepository.server);
    unitTestExecutor = UnitTestExecutor(generationRepository);
  }

  Future<void> saveForClass(
      {required ClassElement classElement,
      required ElementAnnotation annotation,
      required CompilationUnit compilationUnit,
      required ClassDeclaration classAstNode,
      required PubspecData pubspecData,
      required File classFile,
      required TestRunInfo testRunInfo}) async {
    final List<String> excludedMethods = annotation.excludeMethods;
    final fileImport =
        'package:${pubspecData.packageName}/${PathUtils.relatetiveLibFilePath(classFile.path)}';

    final methodAstNodes = await classElement.getMethodDeclarations;
    List<Future> saveFutures = [];
    for (int i = 0; i < methodAstNodes.entries.length; i++) {
      final method = methodAstNodes.entries.elementAt(i);
      saveFutures.add(_saveTestsForMethod(
          method: method,
          compilationUnit: compilationUnit,
          classAstNode: classAstNode,
          fileImport: fileImport,
          excludedMethods: excludedMethods,
          subjectFile: classFile,
          testRunInfo: testRunInfo,
          className: classElement.displayName,
          methodAstNodes: methodAstNodes));
    }
    await Future.wait(saveFutures);
  }

  Future<void> saveForExtension(
      {required ExtensionElement extElement,
      required ElementAnnotation annotation,
      required CompilationUnit compilationUnit,
      required ExtensionDeclaration extAstNode,
      required PubspecData pubspecData,
      required File extFile,
      required TestRunInfo testRunInfo}) async {
    final List<String> excludedMethods = annotation.excludeMethods;
    final fileImport =
        'package:${pubspecData.packageName}/${PathUtils.relatetiveLibFilePath(extFile.path)}';

    final methodAstNodes = await extElement.getMethodDeclarations;

    List<Future> saveFutures = [];
    for (int i = 0; i < methodAstNodes.entries.length; i++) {
      final method = methodAstNodes.entries.elementAt(i);
      saveFutures.add(_saveTestsForMethod(
          method: method,
          compilationUnit: compilationUnit,
          extAstNode: extAstNode,
          fileImport: fileImport,
          excludedMethods: excludedMethods,
          subjectFile: extFile,
          testRunInfo: testRunInfo,
          className: extElement.displayName,
          methodAstNodes: methodAstNodes));
    }
    await Future.wait(saveFutures);
  }

  Future<void> saveForGlobalExecutableElement(
      {required ExecutableElement methodElement,
      required FunctionDeclaration methodAstNode,
      required CompilationUnit compilationUnit,
      required ElementAnnotation annotation,
      required PubspecData pubspecData,
      required File methodFile,
      required TestRunInfo testRunInfo}) async {
    final List<String> excludedMethods = annotation.excludeMethods;
    final fileImport =
        'package:${pubspecData.packageName}/${PathUtils.relatetiveLibFilePath(methodFile.path)}';
    await _saveTestsForMethod(
        method: MapEntry(methodElement.displayName, methodAstNode),
        compilationUnit: compilationUnit,
        fileImport: fileImport,
        excludedMethods: excludedMethods,
        subjectFile: methodFile,
        testRunInfo: testRunInfo);
  }

  Future<void> saveForGlobalMethod(
      {required FunctionElement methodElement,
      required FunctionDeclaration methodAstNode,
      required CompilationUnit compilationUnit,
      required ElementAnnotation annotation,
      required PubspecData pubspecData,
      required File methodFile,
      required TestRunInfo testRunInfo}) async {
    final List<String> excludedMethods = annotation.excludeMethods;
    final fileImport =
        'package:${pubspecData.packageName}/${PathUtils.relatetiveLibFilePath(methodFile.path)}';
    await _saveTestsForMethod(
        method: MapEntry(methodElement.displayName, methodAstNode),
        compilationUnit: compilationUnit,
        fileImport: fileImport,
        excludedMethods: excludedMethods,
        subjectFile: methodFile,
        testRunInfo: testRunInfo);
  }

  Future<void> _saveTestsForMethod({
    required MapEntry<String, Declaration> method,
    required CompilationUnit compilationUnit,
    required String fileImport,
    required List<String> excludedMethods,
    required File subjectFile,
    required TestRunInfo testRunInfo,
    ClassDeclaration? classAstNode,
    ExtensionDeclaration? extAstNode,
    String? className,
    Map<String, MethodDeclaration>? methodAstNodes,
  }) async {
    try {
      final testRequestData = await unitTestExecutor.prepareRequest(
        method: method,
        compilationUnit: compilationUnit,
        classAstNode: classAstNode,
        extensionAstNode: extAstNode,
        fileImport: fileImport,
        className: className,
        methodAstNodes: methodAstNodes,
        excludedMethods: excludedMethods,
      );
      if (testRequestData == null) return;
      testRunInfo.pushFunctionInQueue(method.key);
      final file =
          PathUtils.correspondingPermanentTestFile(method.key, subjectFile);
      if (!(await file.exists())) return;
      final testIsSaved = await cacheManager.checkIfTestIsSaved(
          testRequestData.codeInScrutity(excludeAnnotations: false),
          subjectFile.path,
          testRequestData.functionName,
          className: testRequestData.parentEntityName);
      if (testIsSaved) return; //this test is already saved.
      var hasErrors =
          await fixerExecutor.hasCompilationErrors(file.absolute.path);

      if (hasErrors) {
        wtLog.log(
            "⚠️ File ${PathUtils.relatetiveLibFilePath(file.path)} has errors and will not be updated");
        return;
      }
      final fixedTest = await file.readAsString();
      await unitTestExecutor.updateTest(testRequestData, fixedTest);
      await cacheManager.cacheTest(
          code: testRequestData.codeInScrutity(excludeAnnotations: false),
          codeFilePath: subjectFile.path,
          test: fixedTest,
          testFilePath: file.path,
          functionName: testRequestData.functionName,
          className: testRequestData.parentEntityName,
          saved: true);
      testRunInfo.incrementFromAction(TestGenerationStatus.update, method.key);
    } catch (e, stackTrace) {
      wtTelemetry.trackError(
          severity: Severity.error, error: e, stackTrace: stackTrace);
      wtLog.warning("Error saving unit tests for method: ${method.key}");
      testRunInfo.incrementFromAction(TestGenerationStatus.failed, method.key);
      wtLog.log(e.toString());
    }
  }
}
